#coding: iso-8859-1;
############################################################################
#                                                                          #
#    This file 'SPyRO.py'                                                  #
#    is part of 'SPyRO: Simple Python Remote Objects'                      #
#    Copyright (C) 2004-2005 by Eric Sadit Tellez Avila                    #
#    Copyright (C) 2005-2006 by Eric Sadit Tellez Avila                    #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

import warn
warn = warn.Warn("SPyRO.SPyRO")
import sre
import os
import socket, SocketServer
from threading import Lock
from SocketServer import ThreadingTCPServer
import Queue
import httplib
import time
from shttp import uri
from shttp.uri import URI
import policies
from policies import AuthError, AuthorizationRequest, AuthPasswordRequest
from codepools import CodePool

try:
    import OpenSSL
    import OpenSSL.SSL as SSL
except ImportError, err:
    warn.debug("Unable to use HTTPS as transport layer ", err)
    SSL = None

import args as Args

sleep_min_sec = 0.05
sleep_max_sec = 100.0

SendByReference = 'REF'
SendByValue = 'VAL'

class RemoteException(Exception): pass
class AttributeIsCallable(Exception): pass

import types
from types import DictType

def _set_booleans():
    BooleanType = types.IntType
    __builtins__['True'] = 1
    __builtins__['False'] = 0

try:
    BooleanType = types.BooleanType
    True
except:
    warn.warn("An old version of python or jython was detected")
    _set_booleans()

_HIBRID_SEND_VALUE_tuple = (types.StringType,
                            types.NoneType,
                            types.IntType,
                            BooleanType, 
                            types.LongType, types.FloatType,
                            types.UnicodeType,
                           #types.ComplexType,
                            )
def defaultoverture(httpreq, environment):
    """ By default pickle is disabled, to enable you must overload 'overture' or
    set to None
    
    Any other prelude action must be done in this method. The overture is provided
    to control the execution before start it. The authorities (from policies) are
    checked when a protocol is up and loading. This method must be set to control
    access and tune settings before any protocol and authority.
    """
    q = environment['QUERY_STRING'].lower()
    if 'pickle' in q:
        raise "The pickle marshaler is not allowed on this SPyRO server"

__doc__ = """
Simple Python Remote Objects
Implements Simple Remote Objects (attributes and methods).

Provides the complete suite: server and client services, and RemoteObject
wrappers to objects.

Implement new formats are quite simple. Just create a class or module with
the methods or functions loads and dumps and register your interface
with the |registerFormat| function. Thread safe operation is responsibility
of the new add-on.


Important Note:
Since SPyRO |Server| is multithreaded, but |PeerConnection| had synchronous
protocol, it will not be able to manage many request (every action over a
object is a request) on the same connection. It is necessary use a |Client|
object peer concurrent thread, in the best case, may be one or two are
necessary or use |TClient| (uses a pool of |Client|'s transparentely).
 
"""

import formats
ORBFormats = formats.ORBFormats()
ORBFormats.init()

class TransportProtocolMixIn:
    """ Common methods in the ProtocolHTTP* server and clients"""
    def _choose_format(self, format, request):
        if request and not format:
            format = request.server.defaultFormat
        if format == self.format: return
        self.format = format
        self.broker = ORBFormats.getFormat(format, request)

    def _load_request(self, x):
        # restoring datatypes
        if type(x['auth']) is DictType:
            x['auth'] = AuthorizationRequest().setstate(x['auth'])
        return Request().setstate(x)

    def _load_result(self, x):
        return Response(**x)


class ProtocolHTTPReader(TransportProtocolMixIn):
    """ Protocol used by clients, it runs under HTTP transmition layer """
    def __init__(self, address, location, compress, usingssl=False):
        """ Initialize a client protocol manager.

        Using |address| as a tuple (host,port) and |location|
        Set the remote object location to perform the remote operations.
        In |location| the format of the broker is attached concatenating
        a '?' and a name of the format to the normal location.
        |compress| Use gzip compression in messages
        |usingssl| If True, the connection will be done using https
        
        For example, to request a remote server that is listening on
        /path/SPyRO to use the registered XMLRPC format /path/SPyRO?XMLRPC"""
        self.address = address
        self.usingssl = usingssl
        if not usingssl:
            self.conn = httplib.HTTPConnection(address[0], address[1])
        else:
            self.conn = httplib.HTTPSConnection(address[0], address[1])
        self.location = location
        self.__closed = False
        self.format = None
        self._choose_format(uri.query_string(location), None)
        self.compress = compress

    def __del__(self):
        # if not self.__closed: self.close()
        pass

    def close(self):
        # self.__closed = True
        self.conn.close()

    def __request(self, msg, contenttype = "application/xml"):
        """ Send the message |msg|"""
        headers = {'Content-Type': contenttype}
        if self.compress: headers['Accept-Encoding'] = 'x-gzip'
        # self.send_request(self.location, msg, headers=headers)
        conn = self.conn
        conn.putrequest('POST',self.location)
        conn.putheader('Host', "%s:%s"%(self.address[0],self.address[1]))
        conn.putheader('Content-Type', contenttype)
        conn.putheader('Content-Length', len(msg))
        conn.endheaders()
        conn.send(msg)
    
    def get_result(self, commclient):
        """ Read a message and returns. It uses a specific format """
        msg = self.conn.getresponse().read()
        x = self.broker.load_result(msg, commclient)
        return self._load_result(x)
    def _fixauth(self, msg):
        try:
            msg.auth = msg.auth.__dict__
        except:
            pass
        
    def send_control(self, msg, commclient):
        self._fixauth(msg)
        self.__request(self.broker.dump_control(msg.__dict__, commclient))
    
    def send_setAttr(self, msg, commclient):
        """ send a setAttr request """
        self._fixauth(msg)
        self.__request(self.broker.dump_setAttr(msg.__dict__, commclient))

    def send_getAttr(self, msg, commclient):
        """ send a getAttr request """
        self._fixauth(msg)
        self.__request(self.broker.dump_getAttr(msg.__dict__, commclient))

    def send_callMethod(self, msg, commclient):
        """ send a callMethod request """
        self._fixauth(msg)
        self.__request(self.broker.dump_callMethod(msg.__dict__, commclient))

class ProtocolHTTPShared(TransportProtocolMixIn):
    """ An HTTP Server to manage SPyRO requests, attached to another HTTP server """
    
    def __init__(self):
        """ Perform connection using the opened socket |reqsocket|"""
        self.format = None
        self.broker = None
        self.__closed = False

    def get_request(self, format, peerconn):
        """ Read the request and return it as a Request object instance"""
        env = peerconn.environment
        rfile = peerconn.rfile
        self._choose_format(format, peerconn)
        peerconn.do_next_loop = False
        try:
            x = rfile.read(int(env['CONTENT_LENGTH']))
        except:
            print self.broker,"hola"
            if env['REQUEST_METHOD'] == 'POST':
                x = rfile.read()
            else:
                x = ''
        x = self.broker.load_request(x, peerconn)
        return self._load_request(x)

    def send_data(self, data, peerconn):
        """ Common code between send_result and send_error """
        env = peerconn.environment
        wfile = peerconn.wfile
        wfile.write("Content-Type: %s\r\n"%self.broker.contenttype())
        wfile.write("Content-Length: %s\r\n"%len(data))
        wfile.write("Connection: close\r\n")
        wfile.write("\r\n")
        wfile.write(data)
    
    def send_result(self, msg, peerconn):
        """ Encode and send the message (msg is a Response object) """
        x = self.broker.dump_result(msg, peerconn)
        self.send_data(x, peerconn)

    def send_error(self, errormsg, peerconn, code=None):
        """ Send an error (errormsg is an error string) """
        if code is None: code = Response.ErrorExceptionThrowed
        x = newResponse(code, str(errormsg), SendByValue)
        self.send_data(self.broker.dump_error(x, peerconn), peerconn)

    def close(self):
        """ Close the server """
        if not self.__closed:
            self.__closed = True
    
    def __del__(self):
        if not self.__closed: self.close()

##############################################################################
##############################################################################
#####  The request class
##############################################################################
##############################################################################

class Request:
    """ Object to request to the remote object server

    Values of the variables:
    # Say bye
    ReqBye = 'BYE'
    # Request to Get Remote Attr
    ReqGetAttr = 'GET'
    # Set an attribute
    ReqSetAttr = 'SET'
    # Del an object
    ReqDelObj = 'DEL'
    # Request to Call a remote method
    ReqCallMethod = 'CALL'
    """
    # Say bye
    ReqBye = 'BYE'
    # Request to Get Remote Attr
    ReqGetAttr = 'GET'
    # Set an attribute
    ReqSetAttr = 'SET'
    # Del an object
    ReqDelObj = 'DEL'
    # Request to Call a remote method
    ReqCallMethod = 'CALL'

    def setstate(self, state):
        for k,v in state.items():
            setattr(self,k,v)
        return self

    def __init__(self, objname = None, attrname = None, reqtype = None, auth = None):
        """ A SPyRO Request, It receives:
        |objname| The object name (ID in the server)
        |attrname| The name of the requested attribute
        |reqtype| The type of the request. It can be ReqBye, ReqGetAttr,
        ReqSetAttr, ReqCallMethod """
        self.objname = objname
        self.attrname = attrname
        self.reqtype = reqtype
        self.auth = auth

def newRequestSet(objname, attrname, value, local_server, auth):
    """ Creates a new Request for a |setattr| operation, to set the
    attrname to value"""
    req = Request(objname, attrname, Request.ReqSetAttr, auth)
    req.value = value
    req.local_server = local_server
    return req

def newRequestGet(objname, attrname, rettype, auth):
    """ Creates a new Request for a |getattr| operation """
    r = Request(objname, attrname, Request.ReqGetAttr, auth)
    r.rettype = rettype
    return r

def newRequestDel(objname, auth):
    """ Creates a new Request for a DEL object operation """
    return Request(objname, None, Request.ReqDelObj, auth)

def newRequestBye():
    """ Creates a new Request for a 'YE object operation """
    return Request(None, None, Request.ReqBye, None)

def newRequestCall(objname, attrname, args, kwargs, rettype, local_server,
                   auth, extra):
    """ A SPyRO request, call mode
    |objname| Object name
    |attrname| The methodname
    |args| Positional arguments, tuple
    |kwargs| Keyword arguments, dictionary
    |rettype| The return type, SendByReference or SendByValue
    |local_server| If Request to call a remote method and exists port we call
    the method with the pass the arguments by reference the server receives the
    name or id in the remote peer (the caller) and a port to connect. The IP
    address is obtained by the socket
    |auth| The authentication object
    |extra| Additional info to append to message, for example routed calls
    """
    req = Request(objname, attrname, Request.ReqCallMethod, auth)
    req.args = args
    req.kwargs = kwargs
    req.rettype = rettype
    req.local_server = local_server
    if extra: req.extra = extra
    return req

##############################################################################
##############################################################################
#####  The response class set
##############################################################################
##############################################################################

def ObjectWrapper_GuessType(xobject, xserver):
    """ Function to wrap object guessing the more convenient type to send
    (send by reference or by value) """
    if type(xobject) in _HIBRID_SEND_VALUE_tuple:
        return ObjectWrapper_ByValue(xobject)
    else:
        return ObjectWrapper_ByReference(xobject, xserver)

def ObjectWrapper_ByReference(xobject, xserver, xauthority = None):
    """ Wrap objects to send objects to Reference """
    xobjname = '!' + str(id(xobject))
    xserver.setsend(xobjname, xobject, xauthority)
    return (xobjname, SendByReference)

def ObjectWrapper_ByValue(xobject):
    """ Wrap objects to send objects to Reference """
    return (xobject, SendByValue)

def ObjectWrapper_Unpack(objarg):
    return objarg

def ObjectWrapper_GetObject(objarg):
    return objarg[0]

def ObjectWrapper_GetType(objarg):
    return objarg[1]

def newResponse(errcode, retvalue, rettype):
    """ Create and returns a server side response from the arguments """
    return {'errcode': errcode, 'retvalue' : retvalue, 'rettype': rettype}
    
class Response:
    """ Objects to send as response at requests.
    The request codes are the following:
    Successful = 0
    # Unknown Object, no object registered to share in server, retvalue is None
    ErrorUnknownObject = 1
    # Object doesn't have specified attribute, retvalue = None
    ErrorUnknownAttribute = 2 # NO USAR
    # An Exception Throw, exception is returned in retvalue
    ErrorExceptionThrowed = 3
    # An error on protocol
    ErrorUnknownRequestType = 4
    # The object is callable, we don't wanna send it, we want to call it
    ObjectIsCallable = 5

    """
    # No error, successful call
    Successful = 0
    # Unknown Object, no object registered to share in server, retvalue is None
    ErrorUnknownObject = 1
    # Object doesn't have specified attribute, retvalue = None
    ErrorUnknownAttribute = 2 # NO USAR
    # An Exception Throw, exception is returned in retvalue
    ErrorExceptionThrowed = 3
    # An error on protocol
    ErrorUnknownRequestType = 4
    # The object is callable, we don't wanna send it, we want to call it
    ObjectIsCallable = 5

    ErrorNames = ['Successful', 'ErrorUnknownObject','ErrorUnknownAttribute', 
    'ErrorExceptionThrowed', 'ErrorUnknownRequestType','ObjectIsCallable']

    def __init__(self, errcode = None, retvalue = None, rettype = None):
        """ Initialize the response with an error code  |errcode| and the
        return value |retvalue|"""
        self.errcode = errcode
        self.retvalue = retvalue
        self.rettype = rettype

    def errname(self):
        """ Returns the name of the errcode """
        return Response.ErrorNames[self.errcode]
    
    def prepareToCall(self, remoteobj):
        """ Returns the returned value of calling |methodname| with arguments
        |args|. Can raise remote exceptions or locals due to remote errors """
        #self.attrname = self.retvalue.attrname
        #print "**********> PrepareToCall: ",self.retvalue
        self.attrname = self.retvalue[1] #'attrname'
        #self.remoteobj = remoteobj
        self.comm,self.objname,self.objauth = remoteobj.__xxxconnxxx__
        # preparing cache of callable objects
        remoteobj.__dict__[self.attrname] = self.__call__

    def __call__(self, *args, **kwargs):
        """ Internal function, this is the real __call__ method """
        #comm, objname, auth = self.remoteobj.__xxxconnxxx__
        #return comm.callMethodGuess(objname, self.attrname, args, kwargs,
        #                            SendByValue, comm.local_server,
        #                            auth = auth)
        return self.comm.callMethodGuess(self.objname, self.attrname,
                                         args, kwargs, SendByValue,
                                         self.comm.local_server,
                                         auth = self.objauth)

##############################################################################
##############################################################################
#####  The server class set
##############################################################################
##############################################################################

class PeerConnection:
    """ Connect to Remote Host and create a channel to communicate
    This object must be used in a synchronous way
    """
    def __init__(self):
        # This variables must be set in the handle method
        self.wfile = None # Output file, write only
        self.ifile = None # Input file, read only
        self.environment = None # Environment of the request

    def callObjectErrorHandler(self, obj, request, exception, description):
        """ When an exception is catched while serving objects we call an
        special method in servers named |__spyro_error_handler__|. This
        method will receive:
        __spyro_error_handler__(peerconnection, request, exception,description)
        |peerconnection| The resolver of this connection
        |request| The SPyRO request
        |exception| The exception catched
        |description| A description of the problem."""

        if not hasattr(obj,'__spyro_error_handler__'): return
        error_handler = getattr(obj,'__spyro_error_handler__')
        error_handler(self, request, exception, description)

    def getcomm(self, serv, request):
        """ Returns the CommunicationClient associated to the address that
        made the |request| through the |serv| """
        l = request.local_server
        addr=(self.request.getpeername()[0],l[0])
        comm = serv.commclients.get(addr,None)
        if comm == None:
            comm = CommunicationClientThreaded(addr, l[1],
                                               local_server=serv,poolSize=5)
            serv.commclients[addr] = comm
        return comm

    #def __del__(self):
    #    warn.log("DEL PEERCONNECTION", self.__dict__)
        
    def getargument(self, comm, arg):
        """ Get the argument asociated to |arg| using the |comm|
        It manages references and values """
        argobj, argtype = ObjectWrapper_Unpack(arg)
        if argtype == SendByReference:
            return comm[argobj]
        else:
            return argobj

    def handleCall(self, serv, request, obj, atr):
        """ handles the execution of methods. called by handle.
        If the object has the attribute __spyro__ it must be an object
        with the |__contains__| method and every element is a method
        that must be handled by pass (call methods with the request
        method as the first argument, other arguments are passed as normal)

        If the object has the attribute '__spyro_formats__' it must be a
        dictionary (object with the __getitem__ method) of { methodname: argumentformats }

        Where |methodname| must be the name of methods where the arguments need to be checked
        |argumentformats| is the format as listed in the SPyRO.args module
        """
        methodname = request.attrname
        try:
            hasspyro = methodname in getattr(obj,'__spyro__')
        except AttributeError:
            hasspyro = False
        try:
            fmt_args = getattr(obj,'__spyro_formats__')
            fmtcomp = None
            try:
                comp = fmt_args['!compiled!']
                # print repr(comp),"<","l" * 70, ">"
                try:
                    fmtcomp = comp[methodname]
                except (KeyError, AttributeError):
                    fmtcomp = None
            except KeyError:
                fmt_args['!compiled!'] = {}
            if fmtcomp is None:
                try:
                    fmtcomp = Args.Arguments(fmt_args[methodname])
                    fmt_args['!compiled!'][methodname] = fmtcomp
                except KeyError:
                    fmtcomp = None
            if not(fmtcomp is None):
                fmtcomp.check(request.args, request.kwargs)
        except AttributeError:
            pass
        if hasspyro:
            return atr(self, request, *request.args, **request.kwargs)
        if request.local_server:
            comm = self.getcomm(serv, request)
            args = [] # positional argument
            for arg in request.args:
                args.append(self.getargument(comm, arg))
            request.args = tuple(args)
            kwargs = request.kwargs
            for k,v in kwargs.items(): # keyword arguments
                kwargs[k] = self.getargument(comm, v)
        return atr(*request.args, **request.kwargs)

    def can_del(self):
        if self.authobj == None: return True
        if self.authobj.can_del(self.spyreq.auth, self): return True
        raise AuthError, "Can't del object %s permission denied"%(self.spyreq.objname)

    def can_set(self):
        if self.authobj == None: return True
        if self.authobj.can_set(self.spyreq.auth, self): return True
        raise AuthError, "Can't set attr '%s' in the object '%s'"%(self.spyreq.attrname, self.spyreq.objname)

    def can_follow_objname(self):
        if self.authobj == None: return True
        if self.authobj.can_follow_objname(self.spyreq.auth, self): return True
        raise AuthError, "Can't follow using dot names in the object '%s'"%(self.spyreq.objname)

    def can_get(self):
        if self.authobj == None: return True
        if self.authobj.can_get(self.spyreq.auth, self): return True
        raise AuthError, "Can't get attr '%s' in the object '%s'"%(self.spyreq.attrname, self.spyreq.objname)

    def can_call(self):
        if self.authobj == None: return True
        if self.authobj.can_call(self.spyreq.auth, self): return True
        raise AuthError, "Can't call in the object '%s'"%(self.spyreq.objname)

    def can_register(self):
        if self.authobj == None: return True
        if self.authobj.can_register(self.spyreq.auth, self): return True
        raise AuthError, "Can't register objects"
        
    def solve_request(self, serv, rfile, wfile, environment, httprequest, overture=None, request_rewrite=None):
        """
        Solve one SPyRO request
        |serv| SPyRO Server
        |rfile| Read file
        |wfile| Write file
        |environment| Environment of the Request
        |httprequest| http request object
        Optional arguments:
        |overture| Overture to be executed before anything. If overture throws an exception
                   solve_request will not be executed. Prototype
                   
                   overture(httprequest,environment)

        |request_rewrite| If it's different of None will be executed after the read of the
                         request and the generated dictionary will be sent and perhaps modified.
                         Prototype

                         requestrewrite(requestdict,httprequest,environment)
        """
        if overture: overture(httprequest, environment)
        self.wfile = wfile
        self.rfile = rfile
        self.server = serv
        self.request = httprequest.request
        self.environment = environment
        broker = serv.transportProtocol()
        serv.regconn(self)
        self.obj = obj = None
        self.spyreq = spyreq = None
        self.authobj = None
        do_next = True
        while do_next:
            do_next = False
            try:
                obj = None
                self.authobj = None
                self.spyreq = spyreq = broker.get_request(environment['QUERY_STRING'],self)
                if request_rewrite: request_rewrite(self.spyreq, self, environment)
                # say bye
                if Request.ReqBye == spyreq.reqtype:
                    broker.send_result(newResponse(Response.Successful, None,
                                                   SendByValue), self)
                    break
    
                # Deleting objects (generated by users)
                if Request.ReqDelObj == spyreq.reqtype and self.can_del():
                    if spyreq.objname[0] == '!':
                        warn.debug("-DEL-OBJ-",spyreq.objname)
                        #serv.delobj(spyreq.objname)
                    continue
    
                ##########################
                ### Object name resolution
                ##########################
                try:
                    obj, self.authobj = serv.getobj(spyreq.objname)
                except KeyError:
                    try:
                        if not self.can_follow_objname(): raise AttributeError, "UnknownObject because: Forbidden attribute access"
                        # Resolve object name: attributes of objects
                        objname = spyreq.objname.split('.')
                        obj, self.authobj = serv.getobj(objname.pop(0))
                        for oname in objname:
                            obj = getattr(obj, oname) # authobj se hereda
                    except AttributeError:
                        broker.send_error('Unknown Object: ' + repr(spyreq.objname), self, Response.ErrorUnknownObject)
                        continue
                self.obj = obj
                #############################################
                ### The 'set' action
                #############################################
                if Request.ReqSetAttr == spyreq.reqtype and self.can_set():
                    if spyreq.local_server:
                        spyreq.value=self.get_argument(self.getcomm(serv,
                                                                    spyreq),
                                                       self.value)
                    setattr(obj, spyreq.attrname, spyreq.value)
                    broker.send_result(newResponse(Response.Successful,
                                                   None, SendByValue), self)
                    continue
                val=None
                #############################################
                ### The 'get' action
                #############################################
                
                atr=getattr(obj, spyreq.attrname)
                if Request.ReqGetAttr == spyreq.reqtype and self.can_get():
                    if callable(atr):
                        broker.send_result(newResponse(Response.ObjectIsCallable,
                                                       (spyreq.objname,spyreq.attrname), SendByValue), self)
                        continue
                    #print "x"*30,">", repr(atr)
                    val = atr
                #############################################
                ### The 'call' action
                #############################################
                elif Request.ReqCallMethod == spyreq.reqtype and self.can_call():
                    val = self.handleCall(serv, spyreq, obj, atr)
                else:
                    broker.send_error('Unknown Request Type:' + repr(spyreq.reqtype), self, Response.ErrorUnknownRequestType)
                    continue
                if spyreq.rettype == SendByReference and self.can_register():
                    res = newResponse(Response.Successful,
                                      serv.setsend("?%s"%id(val),val),
                                      SendByReference)
                else:
                    res = newResponse(Response.Successful, val, SendByValue)
                broker.send_result(res, self)
            except AuthError, error:
                broker.send_error(error, self)
            except Exception, error:
                errstr = """Exception catched in SPyRO: '%s'
                Remote Address: %s. Local Address:%s\n%s\n
                """%(error, self.request.getpeername(),
                     self.request.getsockname(),
                     spyreq and spyreq.__dict__ or 'None')
                warn.warn(errstr)
                self.callObjectErrorHandler(obj, spyreq, error, errstr)
                broker.send_error(error, self)
            except:
                warn.warn()
                broker.send_error("UnknownExceptionType", self)
            ## end try
        ## end while
        self.wfile = None
        self.rfile = None
        self.server = None
        self.environment = None
        broker.close()
        serv.delconn(self)


class ServerObjectPool:
    """ The pool of objects interface. This class is a basic
    pool implemented with a dictionary. To create a different
    pool of objects is necessary to create a class that implement
    every method in this interface.
    """
    def __init__(self, objects = None):
        """ Init the ServerObjectPool object with a """
        if objects is None:
            objects = {}
        self.objects = objects
        self.lock = Lock()

    def setobj(self, name, obj, authobj = None):
        """ Register an object |obj| with the ID |name| """
        self.lock.acquire()
        self.objects[name] = (obj,authobj)
        self.lock.release()

    def getobj(self, name):
        """ Get an object and its authorization data, it must raise a
        KeyError exception if |name| is not present in the pool """
        return self.objects[name]

    def delobj(self, name):
        """ Remove an object with name |name|"""
        del self.objects[name]

    def setsend(self, newid, ret, auth = None):
        """ Register an object (to send by reference)"""
        self.setobj(newid, ret, auth)
        return newid

class CommunicationServerBase:
    """ The base of spyro's communication servers """
    def __init__(self, address, location, priorities, defaultFormat,
                 transportProtocol, objectpool):
        """ |address| The address where the Server is listening
            |location| Location or path in the web server
            |priorities| Priorities of the protocols
            |defaultFormat| The default message format where none is
            specified
            |transportProtocol| The class to manage the transport layer
            |objectpool| The object pool manager , if None then
            a new SeverObjectPool instance is used
        """
        self.server_address = address
        self.transportProtocol = transportProtocol
        self.priorities = priorities
        if defaultFormat is None: defaultFormat = ''
        self.defaultFormat = defaultFormat
        self.location = location
        # Open connections
        # self.connections = {} 
        self.commclients = {}
        # Number of registered returns
        self.lockconn = Lock()
        if not objectpool:
            objectpool = ServerObjectPool()
        self.setobj = objectpool.setobj
        self.getobj = objectpool.getobj
        self.delobj = objectpool.delobj
        self.setsend = objectpool.setsend

    def __getstate__(self):
        """ To pickle functions """
        return {}

    def regconn(self, conn):
        """ Register a client connection """
#         self.lockconn.acquire()
#         self.connections[conn] = 1
#         self.lockconn.release()
        pass

    def delconn(self, conn):
        """ Removes a client connection """
        pass
#         self.lockconn.acquire()
#         del self.connections[conn]
#         self.lockconn.release()

    def __setitem__(self,k,v):
        """Register an object. """
        self.setobj(k,v)

    def set(self,k,v):
         """ Register an object |v| with the ID |k|. Use |setobj| instead"""
         warn.stack("This method is deprecated")
         raise "This method is deprecated"
    
    def close(self):
        """ Close server """
        pass

class CommunicationServerShared(CommunicationServerBase):
    def __init__(self, address, location = "/SPyRO?XMLRPC",
                 priorities = None, defaultFormat = None,
                 overture = defaultoverture,
                 objectpool = None, protocolServer = ProtocolHTTPShared):
        """ |address| The address where the Server is listening
            |location| Location or path in the web server
            |priorities| Priorities of the protocols
            |defaultFormat| The default message format where none is
            specified
            |overture| A function to be called before anything in the connection process
            See PeerConnection.solve_request (the overture argument) to more information
            If it's None, the overture will be skipped.
            |objectpool| The object pool that dispatches the connections
            |protocolServer| The class of the Procotol used in the Server side, by default
            ProtocolHTTPShared is Used
        """
        CommunicationServerBase.__init__(self, address, location, priorities,
                                         defaultFormat, protocolServer,
                                         objectpool = objectpool,
                                         overture = overture)
    
    def solve_request(self, httpreq, environment):
        x = PeerConnection()
        x.solve_request(self, httpreq.rfile, httpreq.wfile, environment, httpreq, overture = self.overture)

class CommunicationServer(CommunicationServerBase):
    """ Starts a standalone SPyRO Server """
    def __init__(self, address, location = "/SPyRO?XMLRPC",
                 priorities = None, defaultFormat = None,
                 objectpool = None,
                 protocolServer = ProtocolHTTPShared,
                 **httpoptions):
        """ |address| The address where the Server is listening
            |location| Location or path in the web server
            |priorities| Priorities of the protocols
            |defaultFormat| The default message format when None is
            specified
            |overture| A function to be called before anything in the connection process
            See PeerConnection.solve_request (the overture argument) to more information
            If it's None, the overture will be skipped.
            |protocolServer| The class of the Procotol used in the Server side, by default
            ProtocolHTTPShared is Used
        """
        # This class must start an HTTP server and link to a CommunicationServerShared
        CommunicationServerBase.__init__(self, address, location, priorities,
                                         defaultFormat, protocolServer,
                                         objectpool = objectpool)
        import shttp
        import shttp.http
        http = shttp.http
        self.httpserver = http.HTTPServer(address, **httpoptions)
        self.httpserver.register_urihandler(location.split('?')[0], self.request_handler)

    def serve_forever(self):
        self.httpserver.serve_forever()
        
    def request_handler(self, httpreq, environment):
        """ SPyRO Handler """
        peerconn = PeerConnection()
        peerconn.solve_request(self, httpreq.rfile, httpreq.wfile, environment, httpreq)

    def handle_request(self):
        self.httpserver.handle_request()

    def close(self):
        self.httpserver.close()
    
ServerShared = CommunicationServerShared
Server = CommunicationServer

##############################################################################
##############################################################################
#####  The client class set
##############################################################################
##############################################################################

def default_socket_error_handler(comm, _socket, addr, error):
    """ The default, fool, error handler in Client connections """
    errno, errstr = error
    if errno == 111:
        seconds = sleep_min_sec
        connected = False
        while True:
            try:
                _socket.connect(addr)
                break
            except socket.error, _error:
                if _error[0] != 111:
                    default_socket_error_handler(comm, _socket, addr, _error)
                    break
            # WISH: Try to find another server or port and connect,
            # sleep if we can't find more servers
            warn.debug("SPyRO Client to %s: %s"%(addr, error),
                       "trying in %f seconds"%(seconds))
            time.sleep(seconds)
            if seconds > sleep_max_sec:
                seconds = sleep_min_sec
            else:
                seconds = seconds + seconds
    else:
        warn.warn(error)
        raise error
    warn.debug("Done")

class CommClientMixIn:
    """ The mixin to CommunicationClient{,Threaded}"""
    def get(self, name, auth = None):
        """ Get a local representation of the remote object with ID |name|. Deprecated. Use |getobj| instead"""
        warn.stack("This is not an error, but this method is deprecated")
        return RemoteObject(self, name, auth)

    def __getitem__(self,name):
        """ Get a local representation of the remote object with ID |name|."""
        return RemoteObject(self, name, None)

    def getobj(self, name, auth = None):
        """ Get a local representation of the remote object with ID |name|"""
        return RemoteObject(self, name, auth)

    def getconn(self):
        """ Get a communication object """
        return self
        #raise "Don't use this method (getconn) directly"

    def putconn(self, conn):
        """ Return a communication Object """
        pass
        #raise "Don't use this method (putconn) directly"

    def handleRequest(self, conn, request, send_method):
        """ Common code to getAttr, setAttr and callMethod """
        #conn = self.getconn()
        try:
            send_method(request, conn)
            response = conn.broker.get_result(conn)
        except socket.error:
            warn.debug("Unable to connect remote peer in SPyRO.handleRequest")
            conn._do_bye = False
            self.putconn(conn)
            conn.close()
            raise
        except:
            warn.warn("Unexpected exception in SPyRO.CommunicationClient*")
            conn._do_bye = False
            conn.close()
            self.putconn(conn)
            raise
        self.putconn(conn)
        if response.errcode == Response.Successful:
            if response.rettype == SendByReference:
                response.retvalue = self[response.retvalue]
            return response
        if response.errcode == Response.ObjectIsCallable:
            raise AttributeIsCallable, response
        if response.errcode == Response.ErrorExceptionThrowed:
            #warn.stack("RemoteException", response.retvalue)
            #warn.stack("RemoteException", type(response))
            raise RemoteException, response.retvalue
        raise RuntimeError, "Remote send code %d (%s)\n%s\n"%(response.errcode,
                            response.errname(), request.__dict__)

    def getAttr(self, objname, attrname, rettype = SendByValue,
                auth = None):
        """ Get an attribute |attrname| of the object |objname| with the method
        to send the object |rettype|, the default is SendByValue
        |auth| The authorizing information to access the object
        """
        request=newRequestGet(objname, attrname, rettype, auth)
        conn = self.getconn()
        return self.handleRequest(conn, request, conn.broker.send_getAttr).retvalue

    def setAttr(self, objname, attrname, value, sendtype=SendByValue,
                local_server = None, auth = None):
        """ Set the attribute |attrname| to |value| in the object |objname|.
        The mode to send the value is |sendtype|.
        The |local_server| is used to send by reference
        |auth| The authorizing information to access the object
        """
        if local_server:
            if sendtype == SendByValue:
                value = ObjectWrapper_ByValue(value, sendtype)
            elif sendtype == SendByReference:
                value = ObjectWrapper_ByReference(value, local_server)
            else:
                value = ObjectWrapper_GuessType(value, local_server)
            local_server = self.get_request_local_server(local_server)
        request = newRequestSet(objname, attrname, value, local_server,
                                auth)
        conn = self.getconn()
        return self.handleRequest(conn, request, conn.broker.send_setAttr).retvalue
    
    def byeRequest(self):
        """ Send a bye request """
        conn = self.getconn()
        self.handleRequest(conn, newRequestBye(), conn.broker.send_control)

    def delObject(self, objname, auth = None):
        """ Get an attribute |attrname| of the object |objname| with the method
        to send the object |rettype|, the default is SendByValue.
        |auth| The authorizing information to access the object
        """
        request=newRequestDel(objname, auth)
        conn = self.getconn()
        return self.handleRequest(conn, request, conn.broker.send_control).retvalue

    def callMethod(self, objname, attrname, args = (), kwargs = {},
                   rettype = SendByValue, local_server = None, auth = None, extra = None):
        """ Call a method. Every argument must be a ObjectWrapper
        with all the necessary information.
        |objname| The object's name that has the method
        |attrname| The name of the object
        |args| The positional arguments of the call
        |kwargs| The keyword arguments of the call
        |rettype| The return type of the result
        |local_server| the local server to send arguments by reference
        |auth| The authorizing information to access the object
        |extra| Extra general information to be passed in the message
        """
        request = newRequestCall(objname, attrname, args, kwargs,
                                 rettype, local_server, auth, extra)
        conn = self.getconn()
        return self.handleRequest(conn, request, conn.broker.send_callMethod).retvalue

    def get_request_local_server(self, local_server):
        """ Return a local_server tuple to make the request """
        return (local_server.server_address[1],local_server.location)
    
    def callMethodGuess(self, objname, attrname, args = (), kwargs = {},
                        rettype = SendByValue, local_server = None,
                        auth = None, extra = None):
        """ Call a method, Guessing the best type to send arguments
        |objname| The object's name that has the method
        |attrname| The name of the object
        |args| The positional arguments of the call
        |kwargs| The keyword arguments of the call
        |rettype| The return type of the result
        |local_server| the local server to send arguments by reference
        |auth| The authorizing information to access the object
        |extra| Additional information to be added to the message
        """
        if local_server:
            warn.debug("Objects references as arguments, GC problem")
            xargs = []
            for arg in args:
                xargs.append(ObjectWrapper_GuessType(arg, local_server))
            args = tuple(xargs)
            for k,v in kwargs.items():
                kwargs[k] = ObjectWrapper_GuessType(v, local_server)
            local_server=self.get_request_local_server(local_server)
        return self.callMethod(objname, attrname, args, kwargs, rettype,
                            local_server, auth = auth, extra = extra)
    

class CommunicationClientBase(CommClientMixIn):
    """ A base class to any CommunicationServer client. Represents a
    communication channel
    with the server """
    def __init__(self, serv_addr, path, local_server=None,
                 socket_error_handler = default_socket_error_handler,
                 compress = False, usingssl = False):
        """ Initialize client.
        |serv_addr| The remote server address (host,port)
        |path| Location of the server (to apache or another general purpose
               http servers). It describes the encoder protocol using the
               query_string '?' notation.
        |local_server| If is set to a CommunicationServer, when a method is
               called the arguments are passed by reference and the remote
               server
               will try to connect to this server (port and location).
               The IP address is the used to contact the remote server.
        |socket_error_handler| The error handler, called when an exception
            is arised in the connection, parsing, or interpreting messages.
        |compress| Compress the messages between peers
        |usingssl| If True, every connection is performed with the https protocol
        """
        self.socket_error_handler = socket_error_handler
        self.path = path
        self.local_server = local_server
        self.__closed = False
        self.broker = ProtocolHTTPReader(serv_addr, path, compress, usingssl = usingssl)
        self._do_bye = True
        
    def __getstate__(self):
        """ To pickle functions """
        return {'path': self.path }
    
    def close(self):
        """ Close the client """
        self.__closed = True
        if self._do_bye:
            self.byeRequest()
        self.broker.close()

    def __del__(self):
        """ called for the garbage collection system. Close the object
        if it is not closed yet.  """
        if not self.__closed: self.close()

class CommunicationClient(CommunicationClientBase):
    """ A Client to CommunicationServer. Represents a communication channel
    with the server """
    def __init__(self, serv_addr, path, local_server=None,
                 socket_error_handler = default_socket_error_handler,
                 compress = False,
                 usingssl = False):
        CommunicationClientBase.__init__(self, serv_addr, path, local_server,
                                         socket_error_handler,
                                         compress = compress,
                                         usingssl = usingssl)
        self.connlock = Lock()
        
    def __getstate__(self):
        """ To pickle functions """
        return CommunicationClientBase.__getstate__(self)

    def getconn(self):
        self.connlock.acquire()
        return self

    def putconn(self, conn):
        self.connlock.release()
    
Client = CommunicationClient

class CommunicationClientThreaded(CommClientMixIn):
    """ The thread safe CommunicationClient. It uses several
    CommunicationClient to manage threading """
    def __init__(self, serv_addr, path, local_server = None, poolSize=0,
                 socket_error_handler = default_socket_error_handler,
                 compress = False,
                 usingssl = False):
        """ Create the threaded client.
        |serv_addr| The server address (host,port).
        |path| The path where is listening the remote host, it has the
        protocol using query_string '?' notation.
        |poolSize| The maximum size of the pool of CommunicationClient objects
        |local_server| If is set to a CommunicationServer, when a method is
        called the arguments are passed by reference and the remote server
        will try to connect to this server (port and location).
        The IP address is the used to contact the remote server.
        |socket_error_handler| The socket error handler.
        |compress| If True it uses gzip compression to content
        |usingssl| If True it uses the https protocol to perform connections
        """
        self.queue = Queue.Queue(poolSize)
        self.path = path
        self.local_server = local_server
        self.ehandler = socket_error_handler
        self.address = serv_addr
        self.__closed = False
        self.compress = compress
        self.usingssl = usingssl
    
    def __getstate__(self):
        """ To pickle functions """
        return {'path': self.path, 'address': self.address}
    
    def getconn(self):
        try:
            comm=self.queue.get_nowait()
        except Queue.Empty:
            comm=CommunicationClientBase(self.address, self.path,
                                         self.local_server,
                                         socket_error_handler = self.ehandler,
                                         compress = self.compress,
                                         usingssl = self.usingssl)
        return comm

    def putconn(self, conn):
        try:
            self.queue.put(conn)
        except Queue.Full:
            pass

    def __del__(self):
        if not self.__closed: self.close()
            
    def close(self):
        while not self.__closed:
            try:
                comm = self.queue.get_nowait()
                comm.close()
            except Queue.Empty:
                break
        self.__closed = True

TClient = CommunicationClientThreaded


##############################################################################
##############################################################################
##### Remote Object Class and Helper methods
##############################################################################
##############################################################################

class FakeIterator:
    """ FakeIterator """
    def __init__(self, obj):
        self.obj = obj
        self.pos = -1
        self.len = len(obj) - 1

    def next(self):
        if self.pos < self.len:
            self.pos = self.pos + 1
            return self.obj[self.pos]
        else:
            raise StopIteration

    def __call__(self):
        return self


class RemoteObject:
    """ Local representation of a RemoteObject. The remote object needs to be
    registered in the server. """
    def __init__(self, connection, objectname, auth):
        self.__dict__['__xxxconnxxx__'] = (connection, objectname, auth)

    def __getstate__(self):
        return {}
    
    def __getattr__(self, attrname):
        """ Returns the remote attribute named |attribute| as local object.
        Can raise remote exceptions or locals due remote errors """
        if attrname == '__iter__':
            #return lambda: (self[i] for i in range(0,len(self)))
            return FakeIterator(self)
        conn,objname,auth = self.__xxxconnxxx__
        try:
            return conn.getAttr(objname, attrname, SendByValue,
                                auth = auth)
        except AttributeIsCallable, response:
            r = response.args[0]
            r.prepareToCall(self)
            return r

    def __del__(self):
        conn, objname, auth = self.__xxxconnxxx__
        if objname[0] == '!':
            warn.debug("-USR-DEL-OBJ-",objname)
            conn.delObject(objname, auth)
    
    def __setattr__(self, attrname, value):
        """ __setattr__('name', value) <==> x.name = value. Where |x| is
        a remote object """
        conn,objname, auth = self.__xxxconnxxx__
        return conn.setAttr(objname, attrname, value, SendByValue, None,
                            auth = auth)

    def __repr__(self):
        """ Internal representation """
        return "<SPyRO.RemoteObject name: %s, local id: %s>"%(getobjname(self),
                                                              hex(id(self)))

#     def __eq__(self, arg):
#         comm, objname, auth = self.__xxxconnxxx__
#         return comm.callMethodGuess(objname,  '__eq__', (arg), {},
#                                     SendByValue, None, auth = auth)
        
#     def __cmp__(self, arg):
#         comm, objname, auth = self.__xxxconnxxx__
#         return comm.callMethodGuess(objname, '__cmp__', (arg), {},
#                                     SendByValue, None, auth = auth)

def getobject(address, location, name, local_server = None, poolSize=0):
    """ Can be used as a shortcut when only we need one object in address """
    comm=CommunicationClientThreaded(address, location, local_server, poolSize)
    return comm.get(name)

def getconn(remoteobject):
    """ Return the connection of the remote object"""
    return remoteobject.__xxxconnxxx__[0]

def getobjname(remoteobject):
    """ Returns the name of the object """
    return remoteobject.__xxxconnxxx__[1]

