import os,sys
from threading import Lock
from exec_code import exec_code
import policies

class ItemCodePool: pass

class CodePool:
    """ Register python modules and executes functions in modules.
    The registration of objects must not be used, instead a list of module's path is given.
    The registrarion is supported to serve setsend calls. The getobj method reads modules
    from path. The module is loaded the first time that is requested, next requests uses
    a cached version of the module. If the module changes in the disk, is reloaded. Every
    time that the module is readed it can execute global statments

    """
    def __init__(self, path):
        """ Init the ServerObjectPool object with a """
        self.path = path
        self.objects = {}
        self.cache = {} # name : timestamp, filename
        self.lock = Lock()
        self.gvars = {}
        
    def setobj(self, name, obj, authobj = None):
        """ Register an object |obj| with the ID |name| """
        self.lock.acquire()
        self.objects[name] = (obj,authobj)
        self.lock.release()

    def _import_(self, name, _globals, _locals, fromlist):
        """
        Replace function of __import__ to support
        import modules from the code pool
        """
        # import imp
        try:
            return self.getobj(name)[0]
            # return imp.load_module(name,*imp.find_module(name,self.path))
        except KeyError,e:
            # warn.warn("ERROR IMPORTACION:",name,self.path,"\n",e,'*************************\n')
            return __import__(name,_globals,_locals,fromlist)

    def _reload_(self, mod):
        """ Reload an SPyRO module or a python module """
        try:
            import sys
            sys.modules[mod.__name__] = self.getobj(mod.__name__)[0]
        except KeyError:
            reload(mod)
    
    def getobj(self, name):
        """ Get an object and its authorization data, it must raise a
        KeyError exception if |name| is not present in the pool """
        if name[0] != '!':
            try:
                authobj = self.objects[name][1]
            except (KeyError, IndexError), err:
                authobj = None
            self.loadmodule(name, authobj)
        return self.objects[name]
    
    def loadmodule(self, name, authobj):
        """ Loads a python's module in the object path (self.path)
        The authority object is set to the |_spyro_authentication| global
        variable in the module.
        If it is not present the |authobj| argument is used instead.
        
        """
        if name.startswith('.') or os.path.sep in name: return
        for p in self.path:
            filename = os.path.join(p, name + '.py')
            try:
                mtime = os.stat(filename).st_mtime
            except:
                continue
            try:
                _filename, _mtime = self.cache[name]
                if _filename == filename and _mtime == mtime: return
            except KeyError:
                pass
            lvars = ItemCodePool()
            lvars.__dict__['__builtins__'] = {}
            lvars.__dict__.update(__builtins__)
            lvars.__dict__['__builtins__']['__import__'] = self._import_
            lvars.__dict__['__name__'] = name
            # lvars.__dict__['reload'] = self._reload_
            # Double setobj operation:
            # Work around to prevent recursive import loops
            # first:
            self.setobj(name, lvars, authobj)
            exec_code(filename, self.gvars, lvars)
            self.cache[name] = (filename, mtime)
            try:
                auth = lvars._spyro_authentication
            except AttributeError:
                if authobj == None: # default accessor
                    auth = lvars._spyro_authentication = policies.PublicAllowCall()
                else:
                    auth = authobj
            # second and final
            self.setobj(name, lvars, auth)
            return lvars

    def delobj(self, name):
        """ Remove an object with name |name|"""
        del self.objects[name]
        try:
            del self.cache[name]
        except:
            pass
    
    def setsend(self, newid, ret, auth = None):
        """ Register an object (to send by reference)"""
        self.setobj(newid, ret, auth)
        return newid
