import socket, os, sys
from SocketServer import BaseServer, ThreadingTCPServer, TCPServer
from BaseHTTPServer import HTTPServer as httpbase
from CGIHTTPServer import CGIHTTPRequestHandler, nobody_uid
try:
    from OpenSSL import SSL
except ImportError:
    import warn
    warn.debug("Unable to use the HTTPS protocol")
    SSL = None
import warn
import urllib
import select
import urlparse
import posixpath
import sre

warn = warn.Warn("shttp.http")

def getkwargs(httpreq,environment):
    import cgi
    fields = cgi.FieldStorage(fp=httpreq.rfile,environ=environment)
    kwargs = {}
    for x in fields.keys():
        kwargs[x] = fields.getvalue(x)
    return kwargs

def webmethod(httpreq, environment, method, returntype="text/plain"):
    wfile = httpreq.wfile
    kwargs = getkwargs(httpreq,environment)
    res = str(method(**kwargs))
    wfile.write("Content-Type: " + returntype + "\r\n")
    wfile.write("Content-Length: " + str(len(res)) + "\r\n")
    wfile.write("\r\n")
    wfile.write(res)


def httpcall(realhandler,returntype='text/plain'):
    """ Creates a lexical closure to execute a function with
    query_string and post requests as function arguments
    |realhandler| The handler of the call
    |returntype| The return type of the function to be placed as
    Content-Type header

    Note: If httpcall is used as decorator the declared function
    will be replaced, use regcall of http object as decorator.
    
    """
    def directcall(httpreq, environment):
        webmethod(httpreq, environment, realhandler, returntype)
    return directcall

class HTTPRequestHandler(CGIHTTPRequestHandler):
    def translate_path(self, path, root=None):
        """Translate a /-separated PATH to the local filename syntax.

        Components that mean special things to the local file system
        (e.g. drive or directory names) are ignored.  (XXX They should
        probably be diagnosed.)

        """
        # abandon query parameters
        path = urlparse.urlparse(path)[2]
        path = posixpath.normpath(urllib.unquote(path))
        words = path.split('/')
        words = filter(None, words)
        final = []
        for word in words:
            drive, word = os.path.splitdrive(word)
            head, word = os.path.split(word)
            if word == os.curdir: continue
            if word == os.pardir:
                try:
                    del final[-1]
                except IndexError:
                    pass
            final.append(word)
        sep = os.sep
        return self.server.document_root  + sep + sep.join(final)

    def setup(self):
        self.connection = self.request
        self.rfile = socket._fileobject(self.request, "rb", self.rbufsize)
        self.wfile = socket._fileobject(self.request, "wb", self.wbufsize)
        self.cgi_directories = self.server.cgi_directories

    def is_cgi(self):
        """ Checks if the URI corresponds to a CGI or a Python Handler """
        path = self.path.split('?')[0]
        # warn.debug( path )
        cgi_handlers = self.server.cgi_handlers
        # The order is not directly predectible
        handler = None
        keys = cgi_handlers.keys()
        keys.sort()
        
        for x in keys:
            i = len(x)
            prefix = path[:i]
            # if prefix == x and (not prefix or path[i] == '/'):
            # warn.debug( "i: %s, key: %s, path: %s, prefix: %s"%(i, x, path, prefix) )
            if prefix == x and (len(path) == i or path[i] == '/'):
                self.cgi_info = path[:i], sre.sub(r"^/+","",self.path[i:]), cgi_handlers[x]
                return True
        return CGIHTTPRequestHandler.is_cgi(self)

    def check_script_sanity(self, scriptname, scriptfile):
        """ Returns True if running the script is OK, False otherwise """
        if not os.path.exists(scriptfile):
            self.send_error(404, "No such CGI script (%r)" % scriptname)
            return False
        if not os.path.isfile(scriptfile):
            self.send_error(403, "CGI script is not a plain file (%r)" %
                            scriptname)
            return False
        ispy = self.is_python(scriptname)
        if not ispy:
            if not (self.have_fork or self.have_popen2 or self.have_popen3):
                self.send_error(403, "CGI script is not a Python script (%r)" %
                                scriptname)
                return False
            if not self.is_executable(scriptfile):
                self.send_error(403, "CGI script is not executable (%r)" %
                                scriptname)
                return False
        return True
        
    def get_environ(self):
        """Execute a CGI script."""
        if len(self.cgi_info) == 3:
            dir, rest, handler = self.cgi_info
        else:
            dir, rest = self.cgi_info
            handler = None
        i = rest.rfind('?')
        # warn.debug("REST:",rest, ', PATH: ', self.path)
        if i >= 0:
            rest, query = rest[:i], rest[i+1:]
        else:
            query = ''
        # warn.debug("REST2:",rest, ", QUERY:",query)
        self.query = query
        i = rest.find('/')
        if i >= 0:
            script, rest = rest[:i], rest[i:]
        else:
            script, rest = rest, ''
        # warn.debug("SCRIPT:",script, ", REST:",rest)
        self.script = script
        self.scriptname = scriptname = dir + '/' + script
        self.scriptfile = scriptfile = self.translate_path(scriptname)
        # warn.debug("script:", script, ", scriptname:" + scriptname, ', scriptfile:' + scriptfile)
        
        if handler is None: # CGI
            if not self.check_script_sanity(scriptname, scriptfile):
                raise "Unable to run %s => %s as CGI"%(scriptname, scriptfile)
        # Reference: http://hoohoo.ncsa.uiuc.edu/cgi/env.html
        # XXX Much of the following could be prepared ahead of time!
        env = {}
        length = self.headers.getheader('content-length')
        if length:
            env['CONTENT_LENGTH'] = length
        self.length = length
        if query:
            env['QUERY_STRING'] = query
        env['REQUEST_METHOD'] = self.command

        # useful for spyro and other simple handlers
        # if self.simple_environ: return env
        
        env['SERVER_SOFTWARE'] = self.version_string()
        env['SERVER_NAME'] = self.request.getsockname()[0] #self.server.server_name 
        env['GATEWAY_INTERFACE'] = 'CGI/1.1'
        env['SERVER_PROTOCOL'] = self.protocol_version
        env['SERVER_PORT'] = str(self.server.server_address[1])
        uqrest = urllib.unquote(rest)
        env['PATH_INFO'] = uqrest
        # env['PATH_TRANSLATED'] = 'No-Available' # self.translate_path(uqrest)
        env['SCRIPT_NAME'] = scriptname
        host = self.address_string()
        if host != self.client_address[0]:
            env['REMOTE_HOST'] = host
        env['REMOTE_ADDR'] = self.client_address[0]
        authorization = self.headers.getheader("authorization")
        if authorization:
            authorization = authorization.split()
            if len(authorization) == 2:
                import base64, binascii
                env['AUTH_TYPE'] = authorization[0]
                if authorization[0].lower() == "basic":
                    try:
                        authorization = base64.decodestring(authorization[1])
                    except binascii.Error:
                        pass
                    else:
                        authorization = authorization.split(':')
                        if len(authorization) == 2:
                            env['REMOTE_USER'] = authorization[0]
        # XXX REMOTE_IDENT
        if self.headers.typeheader is None:
            env['CONTENT_TYPE'] = self.headers.type
        else:
            env['CONTENT_TYPE'] = self.headers.typeheader
        accept = []
        for line in self.headers.getallmatchingheaders('accept'):
            if line[:1] in "\t\n\r ":
                accept.append(line.strip())
            else:
                accept = accept + line[7:].split(',')
        env['HTTP_ACCEPT'] = ','.join(accept)
        ua = self.headers.getheader('user-agent')
        if ua:
            env['HTTP_USER_AGENT'] = ua
        co = filter(None, self.headers.getheaders('cookie'))
        if co:
            env['HTTP_COOKIE'] = ', '.join(co)
        # XXX Other HTTP_* headers
        # Since we're setting the env in the parent, provide empty
        # values to override previously set values
        for k in ('QUERY_STRING', 'REMOTE_HOST', 'CONTENT_LENGTH',
                  'HTTP_USER_AGENT', 'HTTP_COOKIE'):
            env.setdefault(k, "")
        return env

    def run_cgi(self):
        """Execute a CGI script."""
        self.scriptname = self.scriptfile = self.query = self.length = None
        env = self.get_environ()
        query = self.query
        scriptname = self.scriptname
        scriptfile = self.scriptfile
        length = self.length
        
        self.send_response(200, "Script output follows")
        decoded_query = query.replace('+', ' ')

        if len(self.cgi_info) == 3:
            handler = self.cgi_info[2]
            handler(self, env)
            return
        
        if self.have_fork and not self.server.usingssl:
            # Unix -- fork as we should
            args = [self.script]
            if '=' not in decoded_query:
                args.append(decoded_query)
            nobody = nobody_uid()
            self.wfile.flush() # Always flush before forking
            pid = os.fork()
            if pid != 0:
                # Parent
                pid, sts = os.waitpid(pid, 0)
                # throw away additional data [see bug #427345]
                while select.select([self.rfile], [], [], 0)[0]:
                    if not self.rfile.read(1):
                        break
                if sts:
                    self.log_error("CGI script exit status %#x", sts)
                return
            # Child
            try:
                try:
                    os.setuid(nobody)
                except os.error:
                    pass
                os.dup2(self.rfile.fileno(), 0)
                os.dup2(self.wfile.fileno(), 1)
                os.execve(scriptfile, args, env)
            except:
                self.server.handle_error(self.request, self.client_address)
                os._exit(127)

        elif self.have_popen2 or self.have_popen3:
            # Windows -- use popen2 or popen3 to create a subprocess
            import shutil
            if self.have_popen3:
                popenx = os.popen3
            else:
                popenx = os.popen2
            cmdline = scriptfile
            if self.is_python(scriptfile):
                interp = sys.executable
                if interp.lower().endswith("w.exe"):
                    # On Windows, use python.exe, not pythonw.exe
                    interp = interp[:-5] + interp[-4:]
                cmdline = "%s -u %s" % (interp, cmdline)
            if '=' not in query and '"' not in query:
                cmdline = '%s "%s"' % (cmdline, query)
            self.log_message("command: %s", cmdline)
            try:
                nbytes = int(length)
            except (TypeError, ValueError):
                nbytes = 0
            files = popenx(cmdline, 'b')
            fi = files[0]
            fo = files[1]
            if self.have_popen3:
                fe = files[2]
            if self.command.lower() == "post" and nbytes > 0:
                data = self.rfile.read(nbytes)
                fi.write(data)
            # throw away additional data [see bug #427345]
            while select.select([self.rfile._sock], [], [], 0)[0]:
                if not self.rfile._sock.recv(1):
                    break
            fi.close()
            shutil.copyfileobj(fo, self.wfile)
            if self.have_popen3:
                errors = fe.read()
                fe.close()
                if errors:
                    self.log_error('%s', errors)
            sts = fo.close()
            if sts:
                self.log_error("CGI script exit status %#x", sts)
            else:
                self.log_message("CGI script exited OK")

        else:
            # Other O.S. -- execute script in this process
            save_argv = sys.argv
            save_stdin = sys.stdin
            save_stdout = sys.stdout
            save_stderr = sys.stderr
            try:
                save_cwd = os.getcwd()
                try:
                    sys.argv = [scriptfile]
                    if '=' not in decoded_query:
                        sys.argv.append(decoded_query)
                    sys.stdout = self.wfile
                    sys.stdin = self.rfile
                    execfile(scriptfile, {"__name__": "__main__"})
                finally:
                    sys.argv = save_argv
                    sys.stdin = save_stdin
                    sys.stdout = save_stdout
                    sys.stderr = save_stderr
                    os.chdir(save_cwd)
            except SystemExit, sts:
                self.log_error("CGI script exit status %s", str(sts))
            else:
                self.log_message("CGI script exited OK")

    def log_message(self, format, *args):
        warn._write("%s - - [%s] %s\n" % (self.address_string(), self.log_date_time_string(), format%args))

class HTTPServer(ThreadingTCPServer,httpbase):
    def __init__(self, server_address, HandlerClass = HTTPRequestHandler,
                 ssl_privatekey_file=None, ssl_certificate_file=None,  ssl_method=None,
                 cgi_handlers=None, cgi_directories=None, document_root="."):
        # ThreadingTCPServer.__init__(self, server_address, HandlerClass)
        ThreadingTCPServer.__init__(self, server_address, HandlerClass)
        
        if cgi_handlers is None:
            self.cgi_handlers = {}
        else:
            self.cgi_handlers = cgi_handlers
        self.document_root = document_root
        # self.cgi_handlers['/hello'] = self.helloworld
        # self.cgi_handlers['/spyro'] = self.spyro
        if cgi_directories is None: cgi_directories = []
        self.cgi_directories = cgi_directories
        self.usingssl = False
        self.process_next_request = True
        if ssl_privatekey_file:
            if ssl_method is None: ssl_method = SSL.SSLv23_METHOD
            self.usingssl = True
            context = SSL.Context(ssl_method)
            # server.pem's location (containing the server private key and
            # the server certificate).
            context.use_privatekey_file(ssl_privatekey_file)
            context.use_certificate_file(ssl_certificate_file)
            self.socket = SSL.Connection(context, socket.socket(self.address_family,
                                                                self.socket_type))
            httpbase.server_bind(self)
            httpbase.server_activate(self)

        # self.spyroserver = SPyRO.CommunicationServerShared(('127.0.0.1',9999))
        # self.spyroserver.setobj("lista",['cero','uno','dos'])
        # self.spyroserver.setobj("tupla",['tcero','tuno','tdos'])

    def register_urihandler(self, uri, handler):
        """ Set an urihandler
        |uri| The URI to be handled
        |handler| a function that handles the event
        Prototype:
            function(httpreq, environment)
            httpreq is an HTTPRequestHandler Object
        """
        self.cgi_handlers[uri] = handler

    def unregister_urihandler(self, uri):
        """Deletes the uri-handler associated to |uri|"""
        del self.cgi_handlers[uri]
    
    # def spyro(self, httpreq, environment):
    #     """ SPyRO handler """
    #     self.spyroserver.solve_request(httpreq, environment)
        
    def helloworld(self, httpreq, environment):
        print environment
        print httpreq.cgi_info
        print httpreq.path
        print "*" * 30
        wfile = httpreq.wfile
        wfile.write("""Content-Type: text/plain

        Hola mundo
        
        Primer programa con script handlers 
        """)

    print "XXXXXXXXXX: regobj incomplete authentication (not available)"
    def regobj(self, uri, object, returntype="text/plain", authority=None):
        def handler(httpreq, environment):
            res = httpreq.cgi_info[1]
            method = object
            for x in res.split('?')[0].split('/'):
                if x == '': continue
                method = getattr(method,x)
            webmethod(httpreq, environment, method, returntype)
        self.register_urihandler(uri, handler)
        return object

    def regcall(self, uri, returntype='text/plain'):
        """ Register uri |uri| to be handled a function
        This method can be used safelly as decorator
        
        Example
        @httpobj.regcall('/selected-uri','text/html')
        def square(x):
            int(x)
            return x*x
        """
        def decorator(function):
            self.register_urihandler(uri, httpcall(function, returntype))
            return function
        return decorator
    
    def serve_forever(self):
        """ Handles requestgs while self.process_next_request is True """
        while self.process_next_request:
            self.handle_request()

    def process_request(self, request, client_address):
        """ Process one request, overloaded to support implement trusted
        SSL under heavy load """
        if self.usingssl:
            TCPServer.process_request(self, request, client_address)
        else:
            ThreadingTCPServer.process_request(self, request, client_address)

    def close(self):
        self.process_next_request = False


def test(HandlerClass = HTTPRequestHandler, ServerClass = HTTPServer):
    server_address = ('', 4433) # (address, port)
    #httpserver = ServerClass(server_address, HandlerClass, ssl_privatekey_file='/home/sadit/ssl/apache.key', ssl_certificate_file='/home/sadit/ssl/apache.crt')
    httpserver = ServerClass(server_address, HandlerClass)
    address = httpserver.socket.getsockname()
    warn.debug( "HTTP(S) running in ", address[0], ":", address[1] )
    @httpcall
    def myhandler(call,calling=None):
        return """
        Call value: %s\n
        Calling value: %s\n
        
        By 3kong
        """%(call,calling)

    @httpserver.regcall('/square')
    def square(x):
        x = int(x)
        return x * x
    class X:
        def holax(self):
            return "Hola chicuelos x"
    class Y:
        def __init__(self):
            self.x = X()

        def statistics(*args, **kwargs):
            print "ALGO:", args, kwargs
            return "OK"
        
        def holay(self):
            return "Hola chicuelos y"

    import SPyRO as spyro
    spyroserver = spyro.CommunicationServerShared(None)
    spyroserver.setobj("TT0",Y())
    def spyrosolver(httpreq, environment):
        def spyroreqrewrite(req, httpreq, environment):
            print req.__dict__
        peerconn = spyro.PeerConnection()
        peerconn.solve_request(spyroserver, httpreq.rfile, httpreq.wfile, environment, httpreq, request_rewrite=spyroreqrewrite)
    httpserver.regobj('/y',Y())
    httpserver.register_urihandler('/call', myhandler)
    httpserver.register_urihandler('/xspyro',spyrosolver)
    httpserver.register_urihandler('/holacall',httpcall(lambda **x: "Hola cuates" + repr(x)))
    httpserver.serve_forever()

if __name__ == '__main__':
    test()

