#coding: iso-8859-1;
############################################################################
#                                                                          #
#    This file 'shttp/sHTTP.py'                                            #
#    is part of 'SPyRO: Simple Python Remote Objects'                      #
#    Copyright (C) 2004-2005 by Eric Sadit Tllez Avila                    #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

import sre
import os
import sys
import warn

import socket, SocketServer
WantReadError = None
WantWriteError = None
try:
    import OpenSSL
    import OpenSSL.SSL
    SSL = OpenSSL.SSL
    WantReadError = SSL.WantReadError
    WantWriteError = SSL.WantWriteError
except ImportError, err:
    warn.debug("Unable to use HTTPS as transport layer ", err)
    WantReadError = OSError
    WantWriteError = OSError
    SSL = None

from threading import Lock
from SocketServer import ThreadingTCPServer
import Queue
import time
warn = warn.Warn("shttp.sHTTP")
import gzip
from cStringIO import StringIO

try:
    socket_SHUT_RDWR = socket.SHUT_RDWR
except:
    socket_SHUT_RDWR = 2

try:
    True
except:
    __builtins__['True'] = 1
    __builtins__['False'] = 0

try:
    from select import select
    _USE_SELECT = True
except:
    _USE_SELECT = False

__doc__ = """
HTTP package provides a Server and Client (minimal) implementations of the HTTP
protocol. Both, server and client, can convert any socket into server or client
respectively. The server obtain handlers to methods (GET, POST, etc) and
help on most commonly used functions """

HTTPNewline = "\r\n"
#HTTPHead=sre.compile(r"^([\S]+):\s*([^\r]+)\s*\n$")
HTTPHead=sre.compile(r"^([\S]+):\s*([^\r]+)\s*$")
HTTPRequest=sre.compile(r"^(\S+)\s+(\S+)\s+([^\r\n]+)")
HTTPVersion="HTTP/1.1"
HTTPBlankline=sre.compile(r"^\s*$")
# responses hash taked from BaseHTTPServer:
# Table mapping response codes to messages; entries have the
# form {code: (shortmessage, longmessage)}.
# See http://www.w3.org/hypertext/WWW/Protocols/HTTP/HTRESP.html
responses = {
    100: ('Continue', 'Request received, please continue'),
    101: ('Switching Protocols',
          'Switching to new protocol; obey Upgrade header'),
    
    200: ('OK', 'Request fulfilled, document follows'),
    201: ('Created', 'Document created, URL follows'),
    202: ('Accepted',
          'Request accepted, processing continues off-line'),
    203: ('Non-Authoritative Information', 'Request fulfilled from cache'),
    204: ('No response', 'Request fulfilled, nothing follows'),
    205: ('Reset Content', 'Clear input form for further input.'),
    206: ('Partial Content', 'Partial content follows.'),
    
    300: ('Multiple Choices',
          'Object has several resources -- see URI list'),
    301: ('Moved Permanently', 'Object moved permanently -- see URI list'),    302: ('Found', 'Object moved temporarily -- see URI list'),
    303: ('See Other', 'Object moved -- see Method and URL list'),
    304: ('Not modified',
          'Document has not changed since given time'),
    305: ('Use Proxy',
          'You must use proxy specified in Location to access this '
          'resource.'),
    307: ('Temporary Redirect',
          'Object moved temporarily -- see URI list'),
   
    400: ('Bad request',
          'Bad request syntax or unsupported method'),
    
    401: ('Unauthorized',
          'No permission -- see authorization schemes'),
    402: ('Payment required',
          'No payment -- see charging schemes'),
    403: ('Forbidden',
          'Request forbidden -- authorization will not help'),
    404: ('Not Found', 'Nothing matches the given URI'),
    405: ('Method Not Allowed',
          'Specified method is invalid for this server.'),
    406: ('Not Acceptable', 'URI not available in preferred format.'),
    407: ('Proxy Authentication Required', 'You must authenticate with '
          'this proxy before proceeding.'),
    408: ('Request Time-out', 'Request timed out; try again later.'),
    409: ('Conflict', 'Request conflict.'),
    410: ('Gone',
          'URI no longer exists and has been permanently removed.'),
    411: ('Length Required', 'Client must specify Content-Length.'),
    412: ('Precondition Failed', 'Precondition in headers is false.'),
    413: ('Request Entity Too Large', 'Entity is too large.'),
    414: ('Request-URI Too Long', 'URI is too long.'),
    415: ('Unsupported Media Type', 'Entity body in unsupported format.'),
    416: ('Requested Range Not Satisfiable',
          'Cannot satisfy request range.'),
    417: ('Expectation Failed',
          'Expect condition could not be satisfied.'),
    500: ('Internal error', 'Server got itself in trouble'),
    501: ('Not Implemented',
          'Server does not support this operation'),
    502: ('Bad Gateway', 'Invalid responses from another server/proxy.'),
    503: ('Service temporarily overloaded',
          'The server cannot process the request due to a high load'),
    504: ('Gateway timeout',
          'The gateway server did not receive a timely response'),
    505: ('HTTP Version not supported', 'Cannot fulfill request.'),
}

HTTPMessage="%s %d %s\r\nContent-Type: text/plain\r\nContent-Length: 0\r\n"

class HTTPStopServer(Exception): pass
class HTTPError(Exception): pass
class HTTPUnexpectedStatus(HTTPError): pass
class HTTPBadRequest(HTTPError): pass
class HTTPNullRequest(HTTPBadRequest): pass
class HTTPBadResponse(HTTPError): pass
class HTTPBadHeader(HTTPError): pass

class HTTPFileNotFound(HTTPError): pass
class HTTPResolverError(HTTPError): pass

class HTTPURIHandler(HTTPError): pass

class HTTPBase:
    """ Common code to HTTP client and HTTP server """
    def close(self):
        """ close the Object """
        self.__closed = True
        self.close_socket()

    def close_socket(self):
        if self.socket is None: return
        try:
            self.fp.close()
            if self.usingssl:
                self.socket.shutdown()
                self.socket.sock_shutdown(socket_SHUT_RDWR)
            else:
                self.socket.shutdown(socket_SHUT_RDWR)
            self.socket.close()
        except Exception, err:
            pass

    def __del__(self):
        if not self.__closed: self.close()
    
    def set_socket(self, commsocket, usingssl=False):
        """ Change the socket that connecs peers """
        if not (self.socket is None): self.close_socket()
        self.usingssl = usingssl
        self.socket = commsocket
        self.input = self.output = self.fp = socket._fileobject(commsocket, "w+b",2048)
        # self.input = self.output = self.fp = commsocket.makefile("w+b",2048)
            
    def __init__(self, commsocket, usingssl=False):
        """ Initialize the HTTPBase instance. |commsocket| is the socket
        that connects peers.

        NOTE:

        Every specialized sub-class must be careful with flushes over
        self.fp and its file objects. The flushes have an important impact in
        performance of applications.
        """
        self.socket = None
        self.set_socket(commsocket, usingssl)
        self.headers = {}
        self.request = None
        self.__closed = None
        
    def __getitem__(self, v):
        """ Get the |v| HTTP header """
        return self.headers[v]

    def _read(self):
        #warn.debug("================================ _READ ")
        length = int(self.headers.get('Content-Length',-1))
        traenc = self.headers.get("Transfer-Encoding")
        if traenc and traenc == "chunked":
            data = ''
            while True:
                line = self._readline()
                chunkedoptions = sre.split("\s*;\s*", line)
                length = int(chunkedoptions[0],16)
                if length == 0: break
                while True:
                    try:
                        data = self.fp.read(length)
                    except WantReadError, err:
                        continue
                    break
                self._readline()
            # Trash
            if _USE_SELECT:
                rlist,wlist,xlist = [self.input.fileno()], [], []
                while _USE_SELECT:
                    r = select(rlist, wlist, xlist, 0.5)
                    r = r[0]
                    if len(r) == 0: break
                    #self.fp.readline(512)
                    while True:
                        try:
                            self.fp.read(1)
                        except WantReadError:
                            continue
                        break
        else:
            data = ''
            while True:
                try:
                    data = self.fp.read(length)
                except WantReadError, err:
                    continue
                break
#             data = ""
#             if length == -1:
#                 _data = None
#                 while True:
#                     warn.debug("READ")
#                     _data = self.fp.read(1)
#                     warn.debug("READED")
#                     if _data == "": break
#                     print "_____________>'%s'"%(_data)
#                     data = data + self.fp.read(1)
#             else:
#                     data = self.fp.read(length)
        return data
    
    def get_content(self,just_file_pointer=False, get_raw_data=False):
        """ Return the content of the document (a POST in the server,
        a response int the client) """
        if get_raw_data: return self._read()
        if self.headers.get('Content-Encoding','').find('gzip') == -1:
            if just_file_pointer: return self.fp
            return self._read()
        g = gzip.GzipFile(filename=None, mode="rb", fileobj = StringIO(self._read()))
        if just_file_pointer: return g
        return g.read()

    def request_error(self, string):
        """ Called when an HTTP Request error ocurrs, in the server
        is called when a response error is present. Raises the
        respective exception in both modes"""
        raise HTTPBadRequest, "HTTP Request Error %s '%s'"%(self.socket.getpeername(),string)

    def header_error(self, string):
        """ Called when a bad header is present"""
        raise HTTPBadHeader, "HTTP Header Error '%s'"%(string)

    def _readline(self,max=512):
        line = ''
        while len(line) < max:
            try:
                line = line + self.fp.read(1)
            except WantReadError,err:
                continue
            try:
                if line[-1] == '\n':
                    if line[-2:] == '\r\n':
                        line = line[:-2]
                    else:
                        line = line[:-1]
                    break
            except IndexError:
                pass
        return line 
#         while True:
#             try:
#                 return self.fp.readline(max)
#             except WantReadError, error:
#                 continue
   
    def read_headers(self):
        """ Read the headers and the (response status|request)"""
        self.headers.clear()
        line = self._readline()
        #line = self.fp.readline(512)
        # warn.debug( "HEAD:",repr(line) )
        resp = HTTPRequest.match(line)
        if not resp:
            if line == '': raise HTTPNullRequest("")
            self.request_error(line)
        self.request = resp.group(1,2,3)
        while True:
            line=self._readline()
            # warn.debug( "Line: ", repr(line) )
            if HTTPBlankline.match(line): break
            m=HTTPHead.match(line)
            if not m:
                self.header_error(line)
            header, value = m.group(1,2)
            if header == 'Cookie' or header == 'Accept':
                h = self.headers.get(header, [])
                h.append(value)
                value = h
            self.headers[header] = value

    def _write(self, data):
        while True:
            try:
                self.fp.write(data)
            except WantWriteError:
                continue
            break
    
    def put_headers(self, headers = None):
        """ Send the headers, including the response|request """
        if headers is None: headers = {}
        if not headers.has_key('Connection'):
            headers['Connection']='Keep-Alive'
        strings = [ ]
        for k,v in headers.items():
            if k == 'Cookie' or k == 'Accept':
                for values in v: strings.append("%s: %s\r\n"%(k,values))
            else:
                strings.append("%s: %s\r\n"%(k,v))
        strings.append("\r\n")
        self._write("".join(strings))
        # warn.debug( "<HEADERS: |%s|>"%("".join(strings) ))
    
    def get_method(self):
        """ What method is used to communicate (i.e. GET, POST, etc)"""
        return self.request[0]

    def get_uri(self):
        """ The URI or requested location """
        return self.request[1]

    def get_httpversion(self):
        """ The version of the HTTP used"""
        return self.request[2]

class HTTP(HTTPBase):
    """ The light weight HTTP Server """
    def __del__(self):
        HTTPBase.__del__(self)
    
    def __init__(self, reqsocket, handlers=None):
        """ Initialize the HTTP server, uses the socket |reqsocket| and
        uses |handlers| (a dictionary methodname -> handlerfunction) as
        handlers to methods. An URI handler can be used inside the method
        handler.
        """
        HTTPBase.__init__(self, reqsocket)
        #self.headers = {'Set-Cookie': []}
        self.handlers = handlers
        self.set_servername('sHTTP server (sHTTP) v0.2')

    # def set_socket(self, commsocket, usingssl=False):
        # warn.stack("\n\n\nxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx set_socket in server ")
        # HTTPBase.set_socket(self, commsocket, usingssl)
    
    def set_servername(self, servername):
        self.servername = servername
        
    def send_response_header(self, httpversion = HTTPVersion, status = 200,
                      status_string = None):
        """ Send the first line with the response of the request
        |httpversion| is the version of the http response
        |status| the reponse status
        |status_string| a human readable status """
        # check if already sent response's header
        if self._done_send_response:
            return
        else:
            self._done_send_response = True
        if not status_string:
            status_string = responses[status][0]
        self._write("%s %s %s\r\n"%(httpversion,status,status_string))
        # print ("%s %s %s\r\n"%(httpversion,status,status_string))
    
    def init_request(self):
        self._done_recv_headers = False
        self._done_recv_response = True # no used yet
        self._done_send_headers = False
        self._done_send_response = False

    def send_headers(self, headers, status = 200):
        self.send_response_header()
        headers['Server'] = self.servername
        # check for headers sent already
        if self._done_send_headers:
            return
        else:
            self._done_send_headers = True
        self.put_headers(headers)
    
    def send_page(self, stringdata, headers=None, status=200):
        """ Send a page with data |stringdata|, headers |headers|, and
        the status |status| (default OK: 200)"""
        # print "xxxxxxxx SEND PAGE>", self.get_httpversion(), self.headers
        if headers == None: headers = {}
        if not headers.has_key('Content-Type'):
            headers['Content-Type'] = 'text/plain'
        x = self.headers.get('Accept-Encoding')
        if x and x.find('gzip') != -1:
            headers['Content-Encoding'] = 'x-gzip'
            string = StringIO()
            gzfile = gzip.GzipFile(filename=None, mode="wb", fileobj = string)
            gzfile.write(stringdata)
            gzfile.flush()
            gzfile.close()
            stringdata = string.getvalue()
        headers['Content-Length'] = len(stringdata)
        self.send_headers(headers,status)
        self._write(stringdata)
        self.fp.flush()
        # print "xxxxxxxx>", self.get_httpversion(), self.headers
        if self.get_httpversion() != 'HTTP/1.1':
            if self.headers.get('Connection','').lower() != 'keep-alive':
                self.close_socket()
                # print "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx>"
                raise HTTPStopServer("Clean stopped, connection close")

    def resolve(self):
        """ Resolve a request using handlers, if there is no handler,
        it responses with 405 error"""
        self.init_request()
        self.read_headers()
        h=self.handlers.get(self.get_method(), None)
        if h:
            h(self)
        else:
            self.send_page(responses[405][1],status=405)
            return
        
    def request_error(self,string):
        """ Request error manager """
        self.send_page(responses[400][1],status=400)
        raise HTTPBadRequest, "Request Error '%s'"%(string)

    def header_error(self,string):
        """ Header error manager """
        self.send_page(responses[400][0],status=400)
        raise HTTPBadHeader, "Header Error '%s'"%(string)

class HTTPClient(HTTPBase):
    """ The HTTP Client """
    def set_socket(self, conn=None, usingssl=False):
        """ Change the socket, to manage connetion errors and https connections """
        if self.socket: self.close_socket()
        conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if usingssl:
            context = SSL.Context(SSL.SSLv23_METHOD)
            conn = SSL.Connection(context, conn)
        conn.connect(self.address)
        HTTPBase.set_socket(self, conn, usingssl)

    def __del__(self):
        HTTPBase.__del__(self)
        
    def __init__(self, address, defaultheaders=None, sockettimeout=None, usingssl=False):
        """ Initialize the HTTP Client, it uses the |address| to connect """
        self.address = address
        HTTPBase.__init__(self, None, usingssl)
        if defaultheaders is None: defaultheaders = {}
        self.defaultheaders = defaultheaders
        try:
            self.socket.settimeout(sockettimeout)
        except Exception, err:
            warn.debug("There is no socket settimeout: ", err)
    
    def request_error(self, string):
        """ Manage the request error as bas response """
        raise HTTPBadResponse, "HTTP Response Error %s '%s'"%(self.socket.getpeername(),repr(string))
        
    def send_request(self, uri, msg='', version=HTTPVersion, method='POST',
                      headers=None):
        """ Send a request to the server """
        if headers is None: headers = {}
        try:
            if self.socket is None: self.set_socket(None)
            self.__send_request(uri, msg, version, method, headers)
        except socket.error, errdata:
            self.set_socket(None)
            self.__send_request(uri, msg, version, method, headers)
        
    def __send_request(self, uri, msg, version, method, headers):
        """ The real send_request, don't use if you do not know what are you
        doing. This is used only by persistant connections """
        if method == 'POST': headers['Content-Length'] = len(msg)
        for k,v in self.defaultheaders.items():
            if not headers.has_key(k): headers[k] = v
        self._write("%s %s %s\r\n"%(method, uri, version))
        #warn.debug("xxxxxxxxxxxxxxxxx>%s %s %s\r\n"%(method, uri, version))
        #warn.stack()
        self.put_headers(headers)
        if method == 'POST':
            self._write(msg)
        self.fp.flush()
        self.read_headers()
        if method == 'HEAD' and self.headers.get('Connection') == 'close':
            self.close_socket()
        if self.request[1][0] != '2':
            raise HTTPUnexpectedStatus,'Unexpected Status %s: %s'%(self.get_response_status(),
                                                                   self.get_response_reason())

    def get_response_status(self):
        return int(self.request[1])

    def get_response_reason(self):
        return self.request[2].rstrip()

    def get_content(self, just_file_pointer=False, get_raw_data=False):
        r = HTTPBase.get_content(self,just_file_pointer, get_raw_data)
        if get_raw_data: return r
        length = int(self.headers.get('Content-Length',-1))
        traenc = self.headers.get("Transfer-Encoding","")
        if not just_file_pointer:
            if self.headers.get("Connection","") == 'close':
                self.close_socket()
            elif not traenc and length < 0:
                self.close_socket()
        return r
