#coding: iso-8859-1;
############################################################################
#                                                                          #
#    This file 'threadutils/PeriodicExecution.py'                          #
#    is part of 'SPyRO: Simple Python Remote Objects'                      #
#    Copyright (C) 2004-2005 by Eric Sadit Téllez Avila                    #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

import time, threading
from threading import Thread
import warn
warn = warn.Warn("PeriodicExecution")

try:
    True
except:
    __builtins__['True'] = 1
    __builtins__['False'] = 0

class PeriodicExecution(Thread):
    """ Performs execution of |periodicfunction| in the time
    |now| + timetobegin, and periodically repites every |timeinterval|. """
    def __init__(self, periodicfunction, timetobegin=0, timeinterval=604800):
        self.periodicfunction = periodicfunction
        self.timetobegin = timetobegin
        self.timeinterval = timeinterval
        Thread.__init__(self)
        self.cond = threading.Condition()
        self.exit = None
        self.start()

    def stop(self):
        self.cond.acquire()
        self.exit = True
        self.cond.notify()
        self.cond.release()
    
    def run(self):
        while True:
            self.cond.acquire()
            if not self.exit: self.cond.wait(self.timetobegin)
            self.cond.release()
            if self.exit: break
            start = time.time()
            try:
                self.periodicfunction()
            except Exception, e:
                warn.warn("PeriodicExecution",e)
            # if the fetch takes more that |timeinterval| sleeps
            # to the next time
            self.timetobegin = self.timeinterval - (time.time() - start) % \
                               self.timeinterval

class PCExecution:
    def __init__(self, fexecution, exec_every = 1, exec_time = None):
        self.__exec_every = exec_every
        self.__changes = 0
        self.__exec_time = exec_time
        self.__fexecution = fexecution
        if exec_time:
            self.__perexe = PeriodicExecution(self.execute,exec_time,exec_time)
        else:
            self.__perexe = None

    def execute(self):
        if self.__changes == 0: return
        self.__fexecution()
        
    def changed(self):
        self.__changes = 1 + self.__changes
        if self.__changes > 0 and self.__exec_every % self.__changes == 0:
            self.__fexecution()
            self.__changes = 0

    def stop(self):
        if self.__perexe: self.__perexe.stop()
